!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!                                                                    
!     Jacobi method for TEST PROBLEM 1                               
!                             parallelized with OpenMP               
!                                                                    
!--------------------------------------------------------------------
!                                                                    
!     Sample Code for Lecture in NR/HEP2 Spring School               
!                                                                    
!                                     Coded by Hirotada Okawa        
!                                                                    
!====================================================================
!   How to compile and use this program in terminal(bash)
!====================================================================
!   $ gfortran -O2 -ffast-math -fopenmp -o j_omp jacobi_openMP.f90
!   $ export OMP_NUM_THREADS=2
!   $ ./j_omp
!--------------------------------------------------------------------
!   OMP_NUM_THREADS : Change the number of cores you want to use.
!--------------------------------------------------------------------
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

module inc_coord
  implicit none

!--------------------------------------------------------------
!     Grid points
!--------------------------------------------------------------
  integer,parameter :: jli=1
  integer,parameter :: jui=100
  integer,parameter :: jlb=jli-1
  integer,parameter :: jub=jui+1

!--------------------------------------------------------------
!     Maximum/Minimum Coordinates (physical)
!--------------------------------------------------------------
  real(8),parameter :: xlower=0.
  real(8),parameter :: xupper=1.
  real(8),parameter :: dx=(xupper-xlower)/dble(jub-jli)
  real(8),parameter :: dxi=1.d0/dx

!--------------------------------------------------------------
!     Array for Coordinates (physical)
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: x

!--------------------------------------------------------------
!     Variables to solve
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: h, hprev

!--------------------------------------------------------------
!     Source term for Poisson equation
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: src

end module inc_coord

program main
  use inc_coord
  implicit none

!--------------------------------------------------------------
!     Definition of parameters
!--------------------------------------------------------------
  integer,parameter :: stepmax=1d8
  real(8),parameter :: errormax=1.d-10
  real(8),parameter :: fpar=1.d0
  real(8),parameter :: omega=1.d0

!--------------------------------------------------------------
!     Definition of temporary variables to use
!--------------------------------------------------------------
  integer :: j, step
  real(8) :: xx, hh, hx, hxx, hhi, h1
  real(8) :: errortmp,vtmp

!--------------------------------------------------------------
!     Output File
!--------------------------------------------------------------
  open(200,file='h_o.dat')


!--------------------------------------------------------------
!     OpenMP threads folk
!--------------------------------------------------------------
!$OMP PARALLEL DEFAULT(SHARED) PRIVATE(j,xx,src)

!--------------------------------------------------------------
!     Initialization
!--------------------------------------------------------------
!$OMP DO
  do j=jlb,jub
     x(j)  = xlower +(dble(j)-0.5d0)*dx    ! Coordinates
     h(j)  = 1.d0                          ! variable to solve
     hprev(j) = h(j)                       ! previous variable
  end do
!$OMP END DO

!**************************************************************
!     Main Loop
!**************************************************************
  do step=0,stepmax

!--------------------------------------------------------------
!     Preserve data of previous step
!--------------------------------------------------------------
!$OMP DO
     do j=jlb,jub
        hprev(j) = h(j)
     end do
!$OMP END DO

!--------------------------------------------------------------
!     Jacobi Method
!--------------------------------------------------------------
!$OMP DO
     do j=jli,jui

!==============================================================
!     Definition of Source term
!==============================================================
        xx   = x(j)
        src(j) = xx**2*12.

        h(j) = 0.5d0*( hprev(j+1) +hprev(j-1) -dx**2*src(j) )
     end do
!$OMP END DO

!==============================================================
!     Impose Boundary Condition
!==============================================================
!$OMP SINGLE
     h(jub)=x(jub)**4      ! Dirichlet Boundary Condition
     h(jlb)=h(jli)         ! Neumann Boundary Condition
!$OMP END SINGLE

!--------------------------------------------------------------
!     Check if values converge
!--------------------------------------------------------------
     errortmp=0.d0
     vtmp=0.d0
!$OMP BARRIER
!$OMP DO REDUCTION(+:vtmp,errortmp)
     do j=jli,jui
        errortmp = errortmp +(h(j)-hprev(j))*(h(j)-hprev(j))*dx**2
        vtmp = vtmp + dx**2
     end do
!$OMP END DO

!$OMP SINGLE
     errortmp = dsqrt(errortmp/vtmp)
!$OMP END SINGLE
     if( (errortmp.le.errormax) .and. (step.gt.1) ) exit

!--------------------------------------------------------------
!     Next Guess(Mix Previous value) just in case
!--------------------------------------------------------------
!$OMP DO
     do j=jlb,jub
        h(j) = fpar*h(j) +(1.d0-fpar)*hprev(j)
     end do
!$OMP END DO

!$OMP SINGLE
     write(*,*) "Step=",step,"Error=",errortmp
!$OMP END SINGLE
  end do

!--------------------------------------------------------------
!     OpenMP threads join
!--------------------------------------------------------------
!$OMP END PARALLEL

!--------------------------------------------------------------
!     Print Data
!--------------------------------------------------------------
  do j=jlb,jub
     write(200,'(4e16.8e2)') x(j),h(j),hprev(j),src(j)
  end do

!--------------------------------------------------------------
!     End of Program
!--------------------------------------------------------------
  write(*,*) "End of Run",errortmp
  close(200)

end program main
